@@ -1,25 +1,87 @@
 # TODO: Consider suggestions at https://stackoverflow.com/questions/35673895/type-hinting-annotation-pep-484-for-numpy-ndarray
 # TODO: Ensure code follows style guide: https://numpydoc.readthedocs.io/en/latest/format.html
+from numba import jit
 import numpy as np
 from scipy.stats import rankdata
+import tqdm
 
-def grade_entropy(X):
+def grade_entropy(X: np.ndarray) -> np.float64:
+    '''
+    Computes grade entropy for a strict product order
+    on the row space points.
+
+    A strict product order on embeddings of the real
+    numbers gives a partial order. On a particular data
+    set it may be desirable to quantify the extent to
+    which there is a total order vs noorder at all.
+
+    Under such an order, two points :math:`\\vec{x}, \\vec{y} \\in \\mathbb{R}^n`
+    hold the relation :math:`\\vec{x} < \\vec{y}` if-and-only-if
+    :math:`x_j < y_j\\ \\forall j \\in \\{1, \\cdots, n\\}`.
+
+    This function computes
+
+    : math::
+        H_g = -\\sum_{i=1}^{m} p \\circ g(x_i) \\ln (p \\circ g(x_i))
+
+    where :math:`p` is a probability distribution over the grades :math:`g`
+    of the point :math:`x_i` among the indexed set of points :math:`i \\in \\{1, \\cdots, m\\}`
+    according to a strict product order relation.
+
+    Parameters
+    ----------
+    X : array-like
+        An m x n data matrix.
+
+    Returns
+    -------
+    entropy : np.float64
+        Grade entropy of product order.
+
+    References
+    ----------
+    .. [1] "Graded Poset.", https://en.wikipedia.org/wiki/Graded_poset
+    .. [2] "Entropy (Information Theory).", https://en.wikipedia.org/wiki/Entropy_(information_theory)
+    .. [3] "Partially Ordered Set.", https://en.wikipedia.org/wiki/Partially_ordered_set
+    .. [4] "Partially Ordered Set.", https://mathworld.wolfram.com/PartiallyOrderedSet.html
+    .. [5] "Product Order.", https://en.wikipedia.org/wiki/Product_order
+
+    Examples
+    --------
+    >>> import numpy as np
+    >>> data = np.arange(100).reshape(10,10)
+    >>> grade_entropy(data)
+    1.0
+
+    .. warning::
+        The underlying tqdm tool that achieves parallel processing will also print a progress bar. This printout can be needlessly obscure in certain environments such as IDLE. The progress bar should print normally from a BASH environment.
+
+    '''
     P = np.zeros((X.shape[0], X.shape[0]))
-    for i in range(X.shape[0]):
+    @jit(nopython=True, fastmath=True) # does fastmath actually improve speed?
+    def grade(q):
+        v = np.zeros(X.shape[0])
         for j in range(X.shape[0]):
-            if np.all(X[i] < X[j]):
-                P[i, j] += 1
+            if np.all(X[j] < X[q]):
+                v[j] += 1
+        return v
+    for i in tqdm.tqdm(range(X.shape[0])):
+        P[:, i] = grade(i)
     bins, counts = np.unique(np.sum(P, axis=0), return_counts=True)
     probs = counts / np.sum(counts)
     entropy = np.sum(probs * np.log(probs)) / (np.log(1/np.sum(counts)))
     return entropy
 
-# TODO: Add Bessel correction
 def minkowski_deviation(x: np.ndarray, order=2) -> np.float64:
     '''
-    Calculates the Minkowski deviation of
-    order p. When the order = 2, it is the
-    same as the standard deviation.
+    Calculates the Minkowski deviation of order p.
+    When the order = 2, it is the same as the standard
+    deviation.
+
+    This function estimates
+
+    : math::
+        \\text{Dev}_p \\left[ X \\right] \\triangleq \\sqrt[p]{\\mathbb{E}\\left[ |X - \\mathbb{E}[X]| \\right]}
 
     Parameters
     ----------
@@ -29,6 +91,14 @@ def minkowski_deviation(x: np.ndarray, order=2) -> np.float64:
     -------
     result : np.float64
 
+    See Also
+    --------
+    np.std : Standard deviation.
+
+    References
+    ----------
+    .. [1] "Standard Deviation.", https://en.wikipedia.org/wiki/Standard_deviation
+
     Examples
     --------
     >>> import numpy as np
@@ -49,6 +119,11 @@ def reflective_correlation(X: np.ndarray) -> np.float64:
     When given an m x 2 data matrix, it is equivalent to the
     reflective correlation coefficient.
 
+    This function estimates
+
+    : math::
+        R_r \\left[ X_1, \\cdots, X_n \\right] = \\frac{\\mathbb{E} \\left[ \\prod_{j=1}^{n} X_j \\right]}{\\prod_{j=1}^{n} \\sqrt[n]{\\mathbb{E}\\left[ |X_n|^n \\right]}}
+
     Parameters
     ----------
     X : array_like
@@ -89,10 +164,15 @@ def pearson_correlation(X: np.ndarray) -> np.float64:
     coefficient. When given an m x 2 data matrix, it is equivalent
     to the Pearson's r correlation coefficient.
 
+    This function estimates
+
+    : math::
+        R_p \\left[ X_1, \\cdots, X_n \\right] = \\frac{\\mathbb{E} \\left[ \\prod_{j=1}^{n} \\left( X_j - \\mathbb{E}[X_j] \\right) \\right]}{\\prod_{j=1}^{n} \\sqrt[n]{\\mathbb{E}\\left[ |X_n- \\mathbb{E}[X_j]|^n \\right]}}
+
     Parameters
     ----------
-        X : array-like
-            The m x n data matrix.
+    X : array-like
+        An m x n data matrix.
 
     Returns
     -------
@@ -119,6 +199,11 @@ def circular_correlation(X: np.ndarray) -> np.float64:
     This function calculates the n-ary circular correlation
     coefficient. When given an m x 2 data matrix, it is equivalent
     to the circular correlation coefficient.
+
+    This function estimates
+
+    : math::
+        R_c \\left[ X_1, \\cdots, X_n \\right] = \\frac{\\mathbb{E} \\left[ \\prod_{j=1}^{n} \\sin \\left( X_j - \\mathbb{E}[X_j] \\right) \\right]}{\\prod_{j=1}^{n} \\sqrt[n]{\\mathbb{E}\\left[ |\\sin \\left( X_j - \\mathbb{E}[X_j] \\right)|^n \\right]}}
     
     Parameters
     ----------
@@ -151,7 +236,10 @@ def spearman_correlation(X: np.ndarray, method='average') -> np.float64:
     coefficient. When given an m x 2 data matrix, it is equivalent
     to the Spearman's Rho correlation coefficient.
 
-    The available data ranking options are directly from `scipy.stats.rankdata`.
+    This function estimates
+
+    : math::
+        R_c \\left[ X_1, \\cdots, X_n \\right] = \\frac{\\mathbb{E} \\left[ \\prod_{j=1}^{n} \\text{rank} \\left( X_j - \\mathbb{E}[X_j] \\right) \\right]}{\\prod_{j=1}^{n} \\sqrt[n]{\\mathbb{E}\\left[ |\\text{rank} \\left( X_j - \\mathbb{E}[X_j] \\right)|^n \\right]}}
     
     Parameters
     ----------
@@ -177,27 +265,45 @@ def spearman_correlation(X: np.ndarray, method='average') -> np.float64:
     -------
         : np.float64
             Spearman's correlation coefficient.
+
+    See Also
+    --------
+    scipy.stats.rankdata
+
+    Notes
+    -----
+    The available data ranking options are directly from `scipy.stats.rankdata`.
             
     References
     ----------
     .. [1] "Spearman%27s_rank_correlation_coefficient", https://en.wikipedia.org/wiki/Spearman%27s_rank_correlation_coefficient
     .. [2] "scipy.stats.rankdata", https://docs.scipy.org/doc/scipy/reference/generated/scipy.stats.rankdata.html
 
-    See Also
+    Examples
     --------
-    scipy.stats.rankdata
+    >>> import numpy as np
+    >>> data = np.arange(100).reshape(10,10)
+    >>> estimators.spearman_correlation(data)
+    0.9999999999999991
     '''
     transform = rankdata(X, axis=0, method=method)
     transform = transform - np.mean(transform, axis=0)
     return reflective_correlation(transform)
 
-# TODO: Implement
 def angular_disimilarity(X: np.ndarray) -> np.float64:
     ''' 
     Computes the multilinear angular disimilarity.
     When given an m x 2 data matrix, it is equivalent
     to the angular distance.
 
+    This function computes
+
+    : math::
+        \\text{angular disimilarity} \\triangleq \\frac{\\theta}{\\pi}
+
+    where : math:`\\theta` is the result of computing the arccosine on
+    the reflective correlation coefficient.
+
     Parameters
     ----------
     X : array_like
@@ -212,6 +318,13 @@ def angular_disimilarity(X: np.ndarray) -> np.float64:
     ----------
     .. [1] "Angular distance", https://en.wikipedia.org/wiki/Cosine_similarity#Angular_distance_and_similarity
     .. [2] "What is the exact and precise definition of an ANGLE?", https://math.stackexchange.com/questions/583066/what-is-the-exact-and-precise-definition-of-an-angle
+
+    Examples
+    --------
+    >>> import numpy as np
+    >>> data = np.arange(100).reshape(10,10)
+    >>> angular_disimilarity(data)
+    0.00981604173368436
     '''
     return np.arccos(reflective_correlation(X)) / np.pi
 
@@ -227,18 +340,32 @@ def correlation_ratio(X: np.ndarray, y: np.ndarray) -> np.float64:
 
     Parameters
     ----------
-        X: array_like
-            m x n data matrix
-        y: array_like
-            m-dimensional vector of class labels
+    X: array_like
+        m x n data matrix
+    y: array_like
+        m-dimensional vector of class labels
+
+    Returns
+    -------
+    : np.float64
+        Correlation ratio score.
+
+    Raises
+    ------
+    NotImplementedError
 
     References
     ----------
     .. [1] "Correlation Ratio", https://en.wikipedia.org/wiki/Correlation_ratio
 
     Examples
-    ----------
+    --------
+    >>> import numpy as np
+    >>> data = np.arange(100).reshape(10,10)
+    >>> y = np.arange(10)
+    >>> correlation_ratio(data, y)
     '''
+    raise NotImplementedError
     global_means = np.mean(X, axis=0)
     classes = set(y)
     partitions_X = [X[y == c] for c in classes]
@@ -261,6 +388,16 @@ def misiak_correlation(x: np.ndarray, y: np.ndarray, X: np.ndarray) -> np.float6
     ----------
         : np.float64
             Misiak correlation score.
+
+    Examples
+    --------
+    >>> import numpy as np
+    >>> np.random.seed(0)
+    >>> x = np.random.normal(size=10)
+    >>> y = np.random.normal(size=10)
+    >>> X = np.random.normal(size=100).reshape(10,10)
+    >>> misiak_correlation(x,y,X)
+    -0.11209570083901074
     '''
     G = np.empty((X.shape[1]+1, X.shape[1]+1))
     G[0, 0] = x @ y
@@ -413,3 +550,15 @@ def wang_zheng_correlation(X: np.ndarray) -> np.float64:
     result = np.linalg.det(result)
     result = 1 - result
     return result
+
+def weak_inner_correlation():
+    '''
+    Raises
+    ------
+    NotImplementedError
+
+    References
+    ----------
+    .. [1] https://arxiv.org/pdf/1904.09542.pdf
+    '''
+    raise NotImplementedError
